/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE

#include <bundle.h>
#include <bundle_internal.h>
#include <aul.h>

#include <sys/socket.h>
#include <stdlib.h>
#include <stdio.h>

#include <glib.h>
#include <gio/gio.h>
#include <gio/gunixfdlist.h>
#include <glib-unix.h>

#include "message_port_log.h"
#include "message_port_common.h"
#include "message_port_remote.h"


#define MAX_PACKAGE_STR_SIZE 512

#define DBUS_RELEASE_NAME_REPLY_RELEASED        1 /* *< Service was released from the given name */
#define DBUS_RELEASE_NAME_REPLY_NON_EXISTENT    2 /* *< The given name does not exist on the bus */
#define DBUS_RELEASE_NAME_REPLY_NOT_OWNER       3 /* *< Service is not an owner of the given name */

#define MAX_RETRY_CNT 10
#define SOCK_PAIR_SENDER 0
#define SOCK_PAIR_RECEIVER 1

static bool _initialized = false;
static GHashTable *__local_port_info;
static GHashTable *__trusted_app_list_hash;
static GHashTable *__callback_info_hash;
static GHashTable *__sender_appid_hash;

typedef struct message_port_pkt {
	int remote_port_name_len;
	char *remote_port_name;
	bool is_bidirection;
	bool is_trusted;
	int data_len;
	unsigned char *data;
} message_port_pkt_s;

typedef struct message_port_callback_info {
	message_port_local_port_info_s *local_info;
	int local_id;
	char *remote_app_id;
	GIOChannel *gio_read;
	int g_src_id;
} message_port_callback_info_s;

typedef struct callback_key_info {
	int local_id;
	message_port_callback_info_s *callback_info;
} callback_key_info_s;

static void __callback_info_free(gpointer data)
{
	message_port_callback_info_s *callback_info = (message_port_callback_info_s *)data;
	GError *error = NULL;
	if (callback_info == NULL)
		return;

	if (callback_info->remote_app_id)
		FREE_AND_NULL(callback_info->remote_app_id);

	if (callback_info->local_info) {
		if (callback_info->local_info->port_name)
				FREE_AND_NULL(callback_info->local_info->port_name);

		FREE_AND_NULL(callback_info->local_info);
	}

	if (callback_info->gio_read != NULL) {
		g_io_channel_shutdown(callback_info->gio_read, TRUE, &error);
		if (error) {
			_LOGE("g_io_channel_shutdown error : %s", error->message);
			g_error_free(error);
		}
		g_io_channel_unref(callback_info->gio_read);
		callback_info->gio_read = NULL;
	}

	if (callback_info->g_src_id != 0) {
		g_source_remove(callback_info->g_src_id);
		callback_info->g_src_id = 0;
	}

	FREE_AND_NULL(callback_info);
}

/* LCOV_EXCL_START */
static void __callback_info_free_by_info(message_port_callback_info_s *callback_info)
{
	GList *callback_info_list = g_hash_table_lookup(__callback_info_hash, GUINT_TO_POINTER(callback_info->local_id));
	GList *find_list;

	if (callback_info_list == NULL)
		return;

	find_list = g_list_find(callback_info_list, callback_info);
	if (find_list == NULL)
		return;

	callback_info_list = g_list_remove_link(callback_info_list, find_list);
	__callback_info_free(callback_info);
	g_list_free(find_list);
}
/* LCOV_EXCL_STOP */

static void __hash_destroy_callback_info(gpointer data)
{
	GList *callback_list = (GList *)data;

	if (callback_list != NULL)
		g_list_free_full(callback_list, __callback_info_free);
}

/* LCOV_EXCL_START */
static void __hash_destory_local_value(gpointer data)
{
	message_port_local_port_info_s *mli = (message_port_local_port_info_s *)data;
	if (mli) {
		if (mli->port_name)
			free(mli->port_name);
		free(mli);
	}
}
/* LCOV_EXCL_STOP */

static bool __initialize(void)
{
	if (!initialized_common) {
		if (!initialize_common())
			return false;
	}

	if (__local_port_info == NULL) {
		__local_port_info = g_hash_table_new_full(g_direct_hash,  g_direct_equal, NULL, __hash_destory_local_value);
		retvm_if(!__local_port_info, false, "fail to create __local_port_info");
	}

	if (__sender_appid_hash == NULL) {
		__sender_appid_hash = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);
		retvm_if(!__sender_appid_hash, false, "fail to create __sender_appid_hash");
	}

	if (__trusted_app_list_hash == NULL) {
		__trusted_app_list_hash = g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);
		retvm_if(!__trusted_app_list_hash, false, "fail to create __trusted_app_list_hash");
	}

	if (__callback_info_hash == NULL) {
		__callback_info_hash = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, __hash_destroy_callback_info);
		retvm_if(!__callback_info_hash, false, "fail to create __callback_info_hash");
	}

	_initialized = true;

	return true;
}

bool is_local_port_registed(const char *local_port, bool trusted, int *local_id, message_port_local_port_info_s **lpi)
{
	GHashTableIter iter;
	gpointer key, value;

	if (__local_port_info == NULL) {
		_LOGI("There is no registed local port");
		return false;
	}

	g_hash_table_iter_init(&iter, __local_port_info);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		message_port_local_port_info_s *mi = (message_port_local_port_info_s *)value;

		if ((mi->is_trusted == trusted) && strcmp(mi->port_name, local_port) == 0) {
			*local_id = mi->local_id;
			if (lpi != NULL)
				*lpi = mi;
			return true;
		}
	}
	return false;
}

static int __get_sender_pid(GDBusConnection *conn, const char *sender_name)
{
	GDBusMessage *msg = NULL;
	GDBusMessage *reply = NULL;
	GError *err = NULL;
	GVariant *body;
	int pid = 0;

	msg = g_dbus_message_new_method_call("org.freedesktop.DBus", "/org/freedesktop/DBus",
			"org.freedesktop.DBus", "GetConnectionUnixProcessID");
	if (!msg) {
		_LOGE("Can't allocate new method call");
		goto out;
	}

	g_dbus_message_set_body(msg, g_variant_new("(s)", sender_name));
	reply = g_dbus_connection_send_message_with_reply_sync(conn, msg,
							G_DBUS_SEND_MESSAGE_FLAGS_NONE, -1, NULL, NULL, &err);

	if (!reply) {
		if (err != NULL) {
			_LOGE("Failed to get pid [%s]", err->message);
			g_error_free(err);
		}
		goto out;
	}

	body = g_dbus_message_get_body(reply);
	g_variant_get(body, "(u)", &pid);

out:
	if (msg)
		g_object_unref(msg);
	if (reply)
		g_object_unref(reply);

	return pid;
}

static message_port_pkt_s *__message_port_recv_raw(int fd)
{
	message_port_pkt_s *pkt = NULL;
	unsigned int nb;

	pkt = (message_port_pkt_s *)calloc(sizeof(message_port_pkt_s), 1);
	if (pkt == NULL) {
		close(fd);
		return NULL;
	}

	if (read_string_from_socket(fd, (char **)&pkt->remote_port_name, &pkt->remote_port_name_len) != MESSAGE_PORT_ERROR_NONE) {
/* LCOV_EXCL_START */
		LOGE("read socket fail: port_name");
		free(pkt->remote_port_name);
		free(pkt);
		return NULL;
/* LCOV_EXCL_STOP */
	}

	if (read_socket(fd, (char *)&pkt->is_bidirection, sizeof(pkt->is_bidirection), &nb) != MESSAGE_PORT_ERROR_NONE) {
/* LCOV_EXCL_START */
		LOGE("read socket fail: is_bidirection");
		free(pkt->remote_port_name);
		free(pkt);
		return NULL;
/* LCOV_EXCL_STOP */
	}

	if (read_socket(fd, (char *)&pkt->is_trusted, sizeof(pkt->is_trusted), &nb) != MESSAGE_PORT_ERROR_NONE) {
/* LCOV_EXCL_START */
		LOGE("read socket fail: is_trusted");
		free(pkt->remote_port_name);
		free(pkt);
		return NULL;
/* LCOV_EXCL_STOP */
	}

	if (read_string_from_socket(fd, (char **)&pkt->data, &pkt->data_len) != MESSAGE_PORT_ERROR_NONE) {
/* LCOV_EXCL_START */
		LOGE("read socket fail: data");
		if (pkt->data)
			free(pkt->data);
		free(pkt->remote_port_name);
		free(pkt);
		return NULL;
/* LCOV_EXCL_STOP */
	}

	return pkt;
}

static bool __validate_callback_info(callback_key_info_s *key_info)
{
	GList *cb_list;

	cb_list = g_hash_table_lookup(__callback_info_hash,
			GUINT_TO_POINTER(key_info->local_id));
	if (cb_list == NULL) {
		_LOGI("local_info : %d is already released", key_info->local_id);
		return false;
	}

	cb_list = g_list_find(cb_list, key_info->callback_info);
	if (cb_list == NULL) {
		_LOGI("local_info : %d is already released", key_info->local_id);
		return false;
	}

	return true;
}

static bool __validate_local_info(callback_key_info_s *key_info)
{
	GList *cb_list;

	cb_list = g_hash_table_lookup(__local_port_info,
			GUINT_TO_POINTER(key_info->local_id));
	if (cb_list == NULL) {
		_LOGI("local_info : %d is already released", key_info->local_id);
		return false;
	}

	return true;
}

static gboolean __socket_request_handler(GIOChannel *gio,
		GIOCondition cond,
		gpointer data)
{
	int fd = 0;
	message_port_callback_info_s *mi = NULL;
	callback_key_info_s *key_info;
	message_port_pkt_s *pkt = NULL;
	message_port_local_port_info_s *local_port_info;
	bundle *kb = NULL;
	bool ret = true;
	bool existed = true;

	key_info = (callback_key_info_s *)data;
	if (key_info == NULL)
		return FALSE;

	message_port_lock_mutex();
	if (__validate_callback_info(key_info) == false) {
		ret = FALSE;
		existed = FALSE;
		message_port_unlock_mutex();
		goto out;
	}

	if (__validate_local_info(key_info) == false) {
		ret = FALSE;
		message_port_unlock_mutex();
		goto out;
	}
	message_port_unlock_mutex();

	mi = key_info->callback_info;

	local_port_info = mi->local_info;
	if (local_port_info == NULL || local_port_info->callback == NULL) {
		_LOGE("Failed to get callback info");
		ret = FALSE;
		goto out;
	}

	if (cond == G_IO_HUP) {
		_LOGI("socket G_IO_HUP");
		ret = FALSE;
		goto out;
	}

	fd = g_io_channel_unix_get_fd(gio);
	if (fd < 0) {
		_LOGE("fail to get fd from io channel");
		ret = FALSE;
		goto out;
	}

	pkt = __message_port_recv_raw(fd);
	if (pkt == NULL) {
		_LOGE("recv error on SOCKET");
		ret = FALSE;
		goto out;
	}

	kb = bundle_decode(pkt->data, pkt->data_len);
	if (!kb) {
		_LOGE("Invalid argument : message");
		ret = FALSE;
		goto out;
	}

	if (pkt->is_bidirection)
		local_port_info->callback(mi->local_id, mi->remote_app_id,
			pkt->remote_port_name, pkt->is_trusted, kb, local_port_info->user_data);
	else
		local_port_info->callback(mi->local_id, mi->remote_app_id,
			NULL, pkt->is_trusted, kb, local_port_info->user_data);

	bundle_free(kb);

out:
	if (pkt) {
		if (pkt->remote_port_name)
			free(pkt->remote_port_name);
		if (pkt->data)
			free(pkt->data);
		free(pkt);
	}

	if (mi && ret == FALSE && existed == TRUE) {
		message_port_lock_mutex();
		__callback_info_free_by_info(mi);
		message_port_unlock_mutex();
	}

	return ret;
}


static void __socket_destroy_handler(gpointer data)
{
	_LOGI("__socket_destroy_handler");
	callback_key_info_s *key_info = (callback_key_info_s *)data;
	free(key_info);
}

static callback_key_info_s *__create_callback_key_info(message_port_callback_info_s *callback_info)
{
	callback_key_info_s *_key_info = (callback_key_info_s *)
		calloc(1, sizeof(callback_key_info_s));

	if (_key_info == NULL) {
		_LOGE("out of memory");
		return NULL;
	}

	_key_info->local_id = callback_info->local_id;
	_key_info->callback_info = callback_info;

	return _key_info;
}

static message_port_callback_info_s *__create_callback_info(message_port_local_port_info_s *mi, char *local_appid)
{
	message_port_local_port_info_s *local_info = NULL;
	message_port_callback_info_s *callback_info = NULL;
	bool ret = true;

	callback_info = (message_port_callback_info_s *)calloc(1, sizeof(message_port_callback_info_s));
	if (callback_info == NULL) {
		_LOGE("out of memory");
		return NULL;
	}

	local_info = (message_port_local_port_info_s *)calloc(1, sizeof(message_port_local_port_info_s));
	if (local_info == NULL) {
		ret = false;
		_LOGE("out of memory");
		goto out;
	}

	callback_info->local_id = mi->local_id;
	callback_info->local_info = local_info;
	callback_info->remote_app_id = strdup(local_appid);
	if (callback_info->remote_app_id == NULL) {
		ret = false;
		_LOGE("out of memory");
		goto out;
	}

	local_info->port_name = strdup(mi->port_name);
	if (local_info->port_name == NULL) {
		ret = false;
		_LOGE("out of memory");
		goto out;
	}

	local_info->callback = mi->callback;
	local_info->is_trusted = mi->is_trusted;
	local_info->local_id = mi->local_id;
	local_info->user_data = mi->user_data;

out:
	if (ret == false) {
		__callback_info_free(callback_info);
		return NULL;
	}

	return callback_info;
}

static bool __callback_info_append(message_port_callback_info_s *callback_info)
{
	GList *callback_info_list = NULL;
	message_port_callback_info_s *head_callback_info;

	callback_info_list = g_hash_table_lookup(__callback_info_hash, GUINT_TO_POINTER(callback_info->local_id));
	if (callback_info_list == NULL) {
		head_callback_info = (message_port_callback_info_s *)calloc(1, sizeof(message_port_callback_info_s));
		if (head_callback_info == NULL) {
			_LOGE("fail to alloc head_callback_info");
			return false;
		}
		head_callback_info->local_id = 0;
		head_callback_info->remote_app_id = NULL;
		head_callback_info->local_info = NULL;
		head_callback_info->gio_read = NULL;
		head_callback_info->g_src_id = 0;
		callback_info_list = g_list_append(callback_info_list, head_callback_info);
		callback_info_list = g_list_append(callback_info_list, callback_info);
		g_hash_table_insert(__callback_info_hash, GUINT_TO_POINTER(callback_info->local_id), callback_info_list);
	} else {
		callback_info_list = g_list_append(callback_info_list, callback_info);
	}

	return true;
}

static void __callback_info_update_user_data(int local_id, message_port_message_cb callback, void *user_data)
{
	GList *callback_info_list;
	GList *iter;
	message_port_callback_info_s *callback_info;

	callback_info_list = g_hash_table_lookup(__callback_info_hash, GUINT_TO_POINTER(local_id));
	if (callback_info_list != NULL) {
		for (iter = callback_info_list; iter != NULL; iter = iter->next) {
			callback_info = (message_port_callback_info_s *)iter->data;
			if (callback_info->local_info != NULL) {
				callback_info->local_info->callback = callback;
				callback_info->local_info->user_data = user_data;
			}
		}
	} else {
		_LOGE("fail to find local_id %d ", local_id);
	}
}

static bool __receive_message(GVariant *parameters, GDBusMethodInvocation *invocation)
{
	char *local_port = NULL;
	char *local_appid = NULL;
	char *remote_appid = NULL;
	char *remote_port = NULL;
	gboolean local_trusted = false;
	gboolean remote_trusted = false;
	gboolean bi_dir = false;
	int len = 0;

	bundle *data = NULL;
	bundle_raw *raw = NULL;
	message_port_local_port_info_s *mi;
	int local_reg_id = 0;
	message_port_callback_info_s *callback_info = NULL;
	callback_key_info_s *key_info;

	char buf[1024];
	GDBusMessage *msg;
	GUnixFDList *fd_list;
	int fd_len;
	int *returned_fds = NULL;
	int fd;
	bool ret = false;
	char *key_appid;

	g_variant_get(parameters, "(&s&sbb&s&sbu&s)", &local_appid, &local_port, &local_trusted, &bi_dir,
			&remote_appid, &remote_port, &remote_trusted, &len, &raw);

	if (!remote_port) {
		_LOGE("Invalid argument : remote_port is NULL");
		goto out;
	}
	if (!remote_appid) {
		_LOGE("Invalid argument : remote_appid is NULL");
		goto out;
	}

	if (!local_appid) {
		_LOGE("Invalid argument : local_appid");
		goto out;
	}

	message_port_lock_mutex();
	if (!is_local_port_registed(remote_port, remote_trusted, &local_reg_id, &mi)) {
		_LOGE("Invalid argument : remote_port:(%s) trusted(%d)", remote_port, remote_trusted);
		message_port_unlock_mutex();
		goto out;
	}

	callback_info = __create_callback_info(mi, local_appid);
	if (callback_info == NULL) {
		message_port_unlock_mutex();
		goto out;
	}
	message_port_unlock_mutex();

	if (!local_port) {
		_LOGE("Invalid argument : local_port");
		goto out;
	}
	if (strcmp(remote_appid, app_id) != 0) {
		_LOGE("Invalid argument : remote_appid (%s)", remote_appid);
		goto out;
	}
	if (strcmp(remote_port, callback_info->local_info->port_name) != 0) {
		_LOGE("Invalid argument : remote_port (%s)", remote_port);
		goto out;
	}
	if (!len) {
		_LOGE("Invalid argument : data_len");
		goto out;
	}
	if (remote_trusted) {
		if (g_hash_table_lookup(__trusted_app_list_hash, (gpointer)local_appid) == NULL) {
			if (!is_preloaded(local_appid, remote_appid)) {
				int ret = check_certificate(local_appid, remote_appid);
				if (ret == MESSAGE_PORT_ERROR_NONE) {
					key_appid = strdup(local_appid);
					if (key_appid)
						g_hash_table_insert(__trusted_app_list_hash, key_appid, "TRUE");
				} else {
					_LOGE("The application (%s) is not signed with the same certificate",
							local_appid);
					goto out;
				}
			}
		}
	}

	data = bundle_decode(raw, len);
	if (!data) {
		_LOGE("Invalid argument : message");
		goto out;
	}

	LOGD("call calback %s", local_appid);
	if (bi_dir)
		callback_info->local_info->callback(callback_info->local_info->local_id,
			local_appid, local_port, local_trusted, data, callback_info->local_info->user_data);
	else
		callback_info->local_info->callback(callback_info->local_info->local_id,
			local_appid, NULL, false, data, callback_info->local_info->user_data);
	bundle_free(data);

	ret = true;

	msg = g_dbus_method_invocation_get_message(invocation);
	fd_list = g_dbus_message_get_unix_fd_list(msg);

	/* When application send message to self fd_list is NULL */
	if (fd_list != NULL) {
		returned_fds = g_unix_fd_list_steal_fds(fd_list, &fd_len);
		if (returned_fds == NULL) {
			_LOGE("fail to get fds");
			ret = false;
			goto out;
		}
		fd = returned_fds[0];

		LOGI("g_unix_fd_list_get %d fd: [%d]", fd_len, fd);
		if (fd > 0) {
			callback_info->gio_read = g_io_channel_unix_new(fd);
			if (!callback_info->gio_read) {
				_LOGE("Error is %s\n", strerror_r(errno, buf, sizeof(buf)));
				ret = false;
				goto out;
			}

			key_info = __create_callback_key_info(callback_info);
			if (key_info == NULL) {
				_LOGE("out of memory");
				ret = false;
				goto out;
			}

			callback_info->g_src_id = g_io_add_watch_full(
					callback_info->gio_read,
					G_PRIORITY_DEFAULT,
					G_IO_IN | G_IO_HUP,
					__socket_request_handler,
					(gpointer)key_info,
					__socket_destroy_handler);
			if (callback_info->g_src_id == 0) {
				_LOGE("fail to add watch on socket");
				free(key_info);
				ret = false;
				goto out;
			}

			message_port_lock_mutex();
			if (__callback_info_append(callback_info) == false) {
				_LOGE("fail to append callback_info");
				ret = false;
			}
			message_port_unlock_mutex();
		}
	}

out:
	if (ret == false)
		__callback_info_free(callback_info);

	if (returned_fds)
		free(returned_fds);

	return ret;
}

static void __on_sender_name_appeared(GDBusConnection *connection,
		const gchar     *name,
		const gchar     *name_owner,
		gpointer         user_data)
{
	_LOGI("sender name appeared : %s", name);
}

static void __on_sender_name_vanished(GDBusConnection *connection,
		const gchar     *name,
		gpointer         user_data)
{
	int *watcher_id = (int *)user_data;
	char *local_appid;

	local_appid = (char *)g_hash_table_lookup(__sender_appid_hash, name);
	if (local_appid) {
		g_hash_table_remove(__trusted_app_list_hash, (gpointer)local_appid);
		g_hash_table_remove(__sender_appid_hash, (gpointer)name);
	}

	if (watcher_id) {
		if (*watcher_id > 0)
			g_bus_unwatch_name(*watcher_id);
		else
			LOGE("Invalid watcher_id %d", *watcher_id);
		free(watcher_id);
	} else {
		LOGE("watcher_id is NULL");
	}
}

static bool __check_sender_validation(GVariant *parameters, const char *sender, GDBusConnection *conn)
{
	int ret = 0;
	char buffer[MAX_PACKAGE_STR_SIZE] = {0, };
	char *local_appid = NULL;
	int pid = __get_sender_pid(conn, sender);
	int *watcher_id = (int *)calloc(1, sizeof(int));
	char *_sender;
	char *_appid;
	retvm_if(!watcher_id, false, "Malloc failed");

	ret = aul_app_get_appid_bypid(pid, buffer, sizeof(buffer));
	if (ret != AUL_R_OK) {
		_LOGE("Failed to get the sender ID: (%s) (%d)", sender, pid);
		free(watcher_id);
		return false;
	}

	g_variant_get_child(parameters, 0, "&s", &local_appid);
	if (local_appid == NULL) {
		_LOGE("appid is NULL : (%s) (%d)", sender, pid);
		free(watcher_id);
		return false;
	}

	if (strncmp(buffer, local_appid, MAX_PACKAGE_STR_SIZE) == 0) {
		_LOGD("insert sender !!!!! %s", sender);
		_sender = strdup(sender);
		if (_sender == NULL) {
			_LOGE("out of memory");
			free(watcher_id);
			return false;
		}

		_appid = strdup(local_appid);
		if (_appid == NULL) {
			_LOGE("out of memory");
			free(watcher_id);
			free(_sender);
			return false;
		}

		g_hash_table_insert(__sender_appid_hash, (gpointer)_sender, (gpointer)_appid);
		*watcher_id = g_bus_watch_name_on_connection(
					gdbus_conn,
					sender,
					G_BUS_NAME_WATCHER_FLAGS_NONE,
					__on_sender_name_appeared,
					__on_sender_name_vanished,
					watcher_id,
					NULL);
	} else {
		free(watcher_id);
		return false;
	}
	return true;
}

static void __dbus_method_call_handler(GDBusConnection *conn,
				const gchar *sender, const gchar *object_path,
				const gchar *iface_name, const gchar *method_name,
				GVariant *parameters, GDBusMethodInvocation *invocation,
				gpointer user_data)
{
	_LOGI("method_name: %s, sender: %s", method_name, sender);
	gpointer sender_appid = g_hash_table_lookup(__sender_appid_hash, sender);
	if (sender_appid == NULL) {
		if (!__check_sender_validation(parameters, sender, conn))
			goto out;
	}
	if (g_strcmp0(method_name, "send_message") == 0)
		__receive_message(parameters, invocation);
out:
	g_dbus_method_invocation_return_value(invocation, NULL);
}

static const GDBusInterfaceVTable interface_vtable = {
	__dbus_method_call_handler,
	NULL,
	NULL
};

static int __register_dbus_interface(const char *port_name, bool is_trusted)
{

	GDBusNodeInfo *introspection_data = NULL;
	int registration_id = 0;

	static gchar introspection_prefix[] =
		"<node>"
		"  <interface name='";

	static gchar introspection_postfix[] =
		"'>"
		"	 <method name='send_message'>"
		"	   <arg type='s' name='local_appid' direction='in'/>"
		"	   <arg type='s' name='local_port' direction='in'/>"
		"	   <arg type='b' name='local_trusted' direction='in'/>"
		"	   <arg type='b' name='bi_dir' direction='in'/>"
		"	   <arg type='s' name='remote_appid' direction='in'/>"
		"	   <arg type='s' name='remote_port' direction='in'/>"
		"	   <arg type='b' name='remote_trusted' direction='in'/>"
		"	   <arg type='u' name='data_len' direction='in'/>"
		"	   <arg type='s' name='data' direction='in'/>"
		"	 </method>"
		"  </interface>"
		"</node>";

	char *introspection_xml = NULL;
	int introspection_xml_len = 0;


	int owner_id = 0;
	GError *error = NULL;
	char *bus_name = NULL;
	char *interface_name = NULL;
	GVariant *result = NULL;

	bus_name = get_encoded_name(app_id, port_name, is_trusted);
	if (!bus_name) {
		_LOGE("Fail to get bus name");
		goto out;
	}
	interface_name = bus_name;

	introspection_xml_len = strlen(introspection_prefix) + strlen(interface_name) +
		strlen(introspection_postfix) + 1;

	introspection_xml = (char *)calloc(introspection_xml_len, sizeof(char));
	if (!introspection_xml) {
		_LOGE("out of memory");
		goto out;
	}


	result = g_dbus_connection_call_sync(
			gdbus_conn,
			DBUS_SERVICE_DBUS,
			DBUS_PATH_DBUS,
			DBUS_INTERFACE_DBUS,
			"RequestName",
			g_variant_new("(su)", bus_name, G_BUS_NAME_OWNER_FLAGS_NONE),
			G_VARIANT_TYPE("(u)"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);
	if (error) {
		_LOGE("RequestName fail : %s", error->message);
		g_error_free(error);
		goto out;
	}
	if (result == NULL) {
		_LOGE("fail to get name NULL");
		goto out;
	}
	g_variant_get(result, "(u)", &owner_id);
	if (owner_id == 0) {
		_LOGE("Acquiring the own name is failed");
		goto out;
	}

	_LOGD("Acquiring the own name : %d", owner_id);

	snprintf(introspection_xml, introspection_xml_len, "%s%s%s", introspection_prefix, interface_name, introspection_postfix);

	introspection_data = g_dbus_node_info_new_for_xml(introspection_xml, NULL);
	if (!introspection_data) {
		_LOGE("g_dbus_node_info_new_for_xml() is failed.");
		goto out;
	}

	registration_id = g_dbus_connection_register_object(gdbus_conn,
						MESSAGEPORT_OBJECT_PATH, introspection_data->interfaces[0],
						&interface_vtable, NULL, NULL, NULL);

	_LOGD("registration_id %d", registration_id);

	if (registration_id == 0) {
		_LOGE("Failed to g_dbus_connection_register_object");
		goto out;
	}

out:
	if (introspection_data)
		g_dbus_node_info_unref(introspection_data);
	if (introspection_xml)
		free(introspection_xml);
	if (bus_name)
		free(bus_name);
	if (result)
		g_variant_unref(result);


	return registration_id;
}

static bool __message_port_register_port(const int local_id, const char *local_port, bool is_trusted, message_port_message_cb callback, void *user_data)
{
	message_port_local_port_info_s *mi = (message_port_local_port_info_s *)calloc(1, sizeof(message_port_local_port_info_s));
	retvm_if(!mi, false, "Malloc failed");

	mi->callback = callback;
	mi->is_trusted = is_trusted;
	mi->port_name = strdup(local_port);
	if (mi->port_name == NULL) {
		_LOGE("Malloc failed (%s)", local_port);
		free(mi);
		return false;
	}
	mi->local_id = local_id;
	mi->user_data = user_data;

	g_hash_table_insert(__local_port_info, GINT_TO_POINTER(mi->local_id), mi);
	return true;
}

int get_local_port_info(int id, message_port_local_port_info_s **info)
{
	message_port_local_port_info_s *mi = (message_port_local_port_info_s *)g_hash_table_lookup(__local_port_info, GINT_TO_POINTER(id));

	if (mi == NULL)
		return MESSAGE_PORT_ERROR_PORT_NOT_FOUND;
	*info = mi;

	return MESSAGE_PORT_ERROR_NONE;
}

int register_message_port(const char *local_port, bool is_trusted, message_port_message_cb callback, void *user_data)
{
	_SECURE_LOGI("local_port : [%s:%s]", local_port, is_trusted ? "trusted" : "non-trusted");

	int local_id = 0;
	message_port_local_port_info_s *port_info;
	if (!_initialized) {
		if (!__initialize())
			return MESSAGE_PORT_ERROR_IO_ERROR;
	}

	/* Check the message port is already registed */
	if (is_local_port_registed(local_port, is_trusted, &local_id, &port_info)) {
		port_info->callback = callback;
		port_info->user_data = user_data;
		__callback_info_update_user_data(local_id, callback, user_data);
		return local_id;
	}

	local_id = __register_dbus_interface(local_port, is_trusted);
	if (local_id < 1) {
		_LOGE("register_dbus_interface fail !!");
		return MESSAGE_PORT_ERROR_OUT_OF_MEMORY;
	}

	if (!__message_port_register_port(local_id, local_port, is_trusted, callback, user_data))
		return MESSAGE_PORT_ERROR_OUT_OF_MEMORY;

	return local_id;
}

int unregister_local_port(int local_port_id, bool trusted_port)
{

	GVariant *result;
	char *bus_name = NULL;
	GError *err = NULL;
	int ret = 0;
	message_port_local_port_info_s *mi;

	_LOGI("unregister : %d", local_port_id);

	if (!_initialized) {
		if (!__initialize())
			return MESSAGE_PORT_ERROR_IO_ERROR;
	}

	mi = (message_port_local_port_info_s *)
		g_hash_table_lookup(__local_port_info, GINT_TO_POINTER(local_port_id));
	if (mi == NULL)
		return MESSAGE_PORT_ERROR_PORT_NOT_FOUND;

	if (mi->is_trusted != trusted_port)
		return MESSAGE_PORT_ERROR_INVALID_PARAMETER;

	g_hash_table_remove(__callback_info_hash, GUINT_TO_POINTER(local_port_id));

	bus_name = get_encoded_name(app_id, mi->port_name, mi->is_trusted);
	if (bus_name == NULL)
		return MESSAGE_PORT_ERROR_OUT_OF_MEMORY;

	g_dbus_connection_unregister_object(gdbus_conn, local_port_id);

	result = g_dbus_connection_call_sync(
			gdbus_conn,
			DBUS_SERVICE_DBUS,
			DBUS_PATH_DBUS,
			DBUS_INTERFACE_DBUS,
			"ReleaseName",
			g_variant_new("(s)", bus_name),
			G_VARIANT_TYPE("(u)"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&err);

	if (bus_name)
		free(bus_name);

	if (err) {
		_LOGE("RequestName fail : %s", err->message);
		g_error_free(err);
		return MESSAGE_PORT_ERROR_PORT_NOT_FOUND;
	}
	g_variant_get(result, "(u)", &ret);

	if (result)
		g_variant_unref(result);

	if (ret != DBUS_RELEASE_NAME_REPLY_RELEASED) {

		if (ret == DBUS_RELEASE_NAME_REPLY_NON_EXISTENT) {
			_LOGE("Port Not exist");
			return MESSAGE_PORT_ERROR_PORT_NOT_FOUND;
		} else if (ret == DBUS_RELEASE_NAME_REPLY_NOT_OWNER) {
			_LOGE("Try to release not owned name. MESSAGE_PORT_ERROR_INVALID_PARAMETER");
			return MESSAGE_PORT_ERROR_INVALID_PARAMETER;
		}
	}
	g_hash_table_remove(__local_port_info, GINT_TO_POINTER(local_port_id));

	return MESSAGE_PORT_ERROR_NONE;
}
